import logging
from pathlib import Path
from typing import Generator, List, Optional

from elbow.extractors import extract_file_meta
from elbow.record import Record, concat
from elbow.sources.filesystem import Crawler
from elbow.typing import StrOrPath

from CPACqc.bids2table.entities import BIDSEntities

from .dataset import extract_dataset
from .metadata import extract_metadata, is_associated_sidecar

logger = logging.getLogger(__name__)


def extract_bids_file(path: StrOrPath, with_meta: bool = True) -> Optional[Record]:
    """
    Extract BIDS entities and metadata from a data file in a BIDS dataset.
    """
    if not is_bids_file(path):
        return None

    try:
        entities = BIDSEntities.from_path(path)
    except (TypeError, ValueError) as exc:
        logger.warning(
            "Incomplete and/or invalid entities in file %s", path, exc_info=exc
        )
        return None

    dset_rec = extract_dataset(path)
    if with_meta:
        meta_rec = extract_metadata(path)
    else:
        meta_rec = Record({"json": None}, types={"json": "json"})
    file_rec = extract_file_meta(path)

    rec = concat({"ds": dset_rec, "ent": entities, "meta": meta_rec, "finfo": file_rec})
    return rec


def extract_bids_subdir(
    path: StrOrPath, exclude: List[str], with_meta: bool = True
) -> Generator[Optional[Record], None, None]:
    """
    Extract BIDS records recursively for all files in a sub-directory.
    """
    for path in Crawler(root=path, skip=exclude, exclude=exclude, follow_links=True):
        yield extract_bids_file(path, with_meta=with_meta)


def is_bids_file(path: StrOrPath) -> bool:
    """
    Check if `path` is a valid BIDS data file. E.g. not a directory or JSON sidecar
    associated to another data file.
    """
    # TODO: other checks?
    #   - skip files matching patterns in .bidsignore?
    path = Path(path)
    return (
        not path.is_dir()
        and path.name.startswith("sub-")
        and not is_associated_sidecar(path)
    )



from CPACqc.main import main
import os
import shutil
from colorama import Fore, Style
import pkg_resources
from CPACqc import __version__  # Import the version number
import argparse
import pandas as pd

class StoreTrueOrString(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        if values is None:
            setattr(namespace, self.dest, True)
        else:
            setattr(namespace, self.dest, values)

def run():

    # Parse command-line arguments
    parser = argparse.ArgumentParser(description="Process BIDS directory and generate QC plots.")
    parser.add_argument("-d", "--bids_dir", required=True, help="Path to the BIDS directory")
    parser.add_argument("-o", "--qc_dir", required=False, help="Path to the QC output directory")
    parser.add_argument("-c", "--config", required=False, help="Config file")
    parser.add_argument("-s", "--sub", nargs='+', required=False, help="Specify subject/participant label(s) to process")
    parser.add_argument("-n", "--n_procs", type=int, default=8, help="Number of processes to use for multiprocessing")
    parser.add_argument("-v", "--version", action='version', version=f'%(prog)s {__version__}', help="Show the version number and exit")
    parser.add_argument("-pdf", "--pdf", nargs='?', const=True, action=StoreTrueOrString, default=True, help="Generate PDF report (default: 'report')")
    parser.add_argument("-html", "--html", action='store_true', help="Generate HTML report")
    
    args = parser.parse_args()

    if args.bids_dir is None:
        print(Fore.RED + "Please specify the BIDS directory.")
        print(Style.RESET_ALL)
        return
    
    if args.qc_dir is None:
        args.qc_dir = os.path.join(os.getcwd(), '.temp_qc')
        print(Fore.YELLOW + f"Output directory not specified. Saving output to {args.qc_dir}")
        print(Style.RESET_ALL)

    if args.pdf is True:
        args.pdf = "report"

    try:
        # Create the QC output directory if it doesn't exist
        os.makedirs(args.qc_dir, exist_ok=True)

        if args.html:
            # Locate the templates directory within the package
            templates_dir = pkg_resources.resource_filename('CPACqc', 'templates')

            # Copy only the index.html file from the templates directory to the QC output directory
            src_file = os.path.join(templates_dir, 'index.html')
            dest_file = os.path.join(args.qc_dir, 'index.html')
            shutil.copy2(src_file, dest_file)

    except Exception as e:
        print(f"Error !! : {e}")
        return  # Exit the function if an error occurs

    not_plotted = main(args.bids_dir, args.qc_dir, args.config, args.sub, args.n_procs, args.pdf)

    if not args.html:
        # remove the qc_dir if not generating HTML report
        print(Fore.YELLOW + f"Removing the QC output directory: {args.qc_dir}")
        print(Style.RESET_ALL)
        shutil.rmtree(args.qc_dir)
    else:
        # combine all the csvs inside qc_dir/csv into one csv and name it results.csv
        csv_dir = os.path.join(args.qc_dir, 'csv')
        if os.path.exists(csv_dir):
            # Combine all the csv files into one
            combined_csv = os.path.join(args.qc_dir, 'results.csv')
            combined_df = pd.concat([pd.read_csv(os.path.join(csv_dir, f)) for f in os.listdir(csv_dir) if f.endswith('.csv')])
            combined_df.to_csv(combined_csv, index=False)
        else:
            print(Fore.RED + "No CSV files found in the QC output directory. Please check the log for details.")
            print(Style.RESET_ALL)
            return
        # Rename the qc_dir to results
        new_qc_dir = os.path.join(os.getcwd(), 'results')
        print(Fore.YELLOW + f"Creating HTML report in results dir: {new_qc_dir}")
        print(Style.RESET_ALL)
        
        # Check if the results directory already exists and remove it if it does
        if os.path.exists(new_qc_dir):
            shutil.rmtree(new_qc_dir)
        
        shutil.move(args.qc_dir, new_qc_dir)
        print(Fore.YELLOW + "Done.")

    if len(not_plotted) > 0:
        print(Fore.RED + "Some files were not plotted. Please check the log for details.")
    else:
        print(Fore.GREEN + "All files were successfully plotted.")
    print(Style.RESET_ALL)


import json
import logging
import traceback
from pathlib import Path

from elbow.record import Record
from elbow.typing import StrOrPath

from CPACqc.bids2table.entities import parse_bids_entities

from .inheritance import _glob, find_bids_parents

logger = logging.getLogger(__name__)


def extract_metadata(path: StrOrPath) -> Record:
    """
    Load the JSON sidecar metadata associated with ``path``. Supports metadata
    inheritance by searching up the directory tree for matching JSON files.
    """
    entities = parse_bids_entities(path)
    query = dict(entities, ext=".json")

    metadata = {}
    sidecars = reversed(list(find_bids_parents(query, start=Path(path).parent)))
    for path in sidecars:
        with open(path) as f:
            try:
                metadata.update(json.load(f))
            except (json.JSONDecodeError, TypeError):
                logger.warning(
                    f"Bad JSON sidecar data {path}\n\n" + traceback.format_exc()
                )

    # TODO: type aliases for json, pickle, etc so we can use a dataclass here.
    rec = Record({"json": metadata or None}, types={"json": "json"})
    return rec


def is_associated_sidecar(path: StrOrPath) -> bool:
    """
    Check if a file is a JSON sidecar associated with other data file(s).
    """
    path = Path(path)

    # Must be JSON
    if not path.suffix == ".json":
        return False

    entities = parse_bids_entities(path)

    # Assume all JSON above the lowest level of hierarchy are associated
    if entities.get("datatype") is None:
        return True

    # All sidecars must contain a suffix
    suffix = entities.get("suffix")
    if suffix is None:
        return False

    # Finally, check if there are any matches at the lowest level
    # If not, we are a key-value file or solo sidecar like an MRIQC IQM JSON.
    # Note this pattern always matches the file itself, so we check if there are any
    # extra matches.
    if len(_glob(path.parent, f"*_{suffix}.*")) > 1:
        return True

    return False



from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

import pandas as pd

from CPACqc.bids2table.entities import ENTITY_NAMES_TO_KEYS, BIDSEntities
from CPACqc.bids2table.extractors.metadata import extract_metadata


class BIDSTable(pd.DataFrame):
    """
    A table representing one or more BIDS datasets.
    """

    @cached_property
    def ds(self) -> pd.DataFrame:
        """
        The dataset (`ds`) subtable.
        """
        return self.nested["ds"]

    @cached_property
    def ent(self) -> pd.DataFrame:
        """
        The entities (`ent`) subtable.
        """
        return self.nested["ent"]

    @cached_property
    def meta(self) -> pd.DataFrame:
        """
        The metadata (`meta`) subtable.
        """
        return self.nested["meta"]

    @cached_property
    def finfo(self) -> pd.DataFrame:
        """
        The file info (`finfo`) subtable.
        """
        return self.nested["finfo"]

    @cached_property
    def files(self) -> List["BIDSFile"]:
        """
        Convert the table to a list of structured `BIDSFile`s.
        """

        def to_dict(val):
            if pd.isna(val):
                return {}
            return dict(val)

        return [
            BIDSFile(
                dataset=row["ds"]["dataset"],
                root=Path(row["ds"]["dataset_path"]),
                path=Path(row["finfo"]["file_path"]),
                entities=BIDSEntities.from_dict(row["ent"]),
                metadata=to_dict(row["meta"]["json"]),
            )
            for _, row in self.nested.iterrows()
        ]

    @cached_property
    def datatypes(self) -> List[str]:
        """
        Get all datatypes present in the table.
        """
        return self.ent["datatype"].unique().tolist()

    @cached_property
    def modalities(self) -> List[str]:
        """
        Get all modalities present in the table.
        """
        # TODO: Is this the right way to get the modality
        return self.ent["mod"].unique().tolist()

    @cached_property
    def subjects(self) -> List[str]:
        """
        Get all unique subjects in the table.
        """
        return self.ent["sub"].unique().tolist()

    @cached_property
    def entities(self) -> List[str]:
        """
        Get all entity keys with at least one non-NA entry in the table.
        """
        entities = self.ent.dropna(axis=1, how="all").columns.tolist()
        special = set(BIDSEntities.special())
        return [key for key in entities if key not in special]

    @cached_property
    def flat_meta(self) -> pd.DataFrame:
        """
        A table of flattened JSON metadata where each metadata field is converted to its
        own column, with nested levels separated by `'.'`.

        See also:

        - [`pd.json_normalize`](https://pandas.pydata.org/docs/reference/api/pandas.json_normalize.html):
        more general function in pandas.
        """
        # Need to replace None with empty dict for max_level=0 to work.
        metadata = pd.json_normalize(
            self["meta__json"].map(lambda v: v or {}), max_level=0
        )
        metadata.index = self.index
        return metadata

    @cached_property
    def nested(self) -> pd.DataFrame:
        """
        A copy of the table with column labels organized in a nested
        [`MultiIndex`](https://pandas.pydata.org/docs/user_guide/advanced.html#hierarchical-indexing-multiindex).
        """
        # Cast back to the base class since we no longer have the full BIDS table
        # structure.
        return pd.DataFrame(flat_to_multi_columns(self))

    @cached_property
    def flat(self) -> pd.DataFrame:
        """
        A copy of the table with subtable prefixes e.g. `ds__`, `ent__` removed.
        """
        return self.nested.droplevel(0, axis=1)

    def filter(
        self,
        key: str,
        value: Optional[Any] = None,
        *,
        items: Optional[Iterable[Any]] = None,
        contains: Optional[str] = None,
        regex: Optional[str] = None,
        func: Optional[Callable[[Any], bool]] = None,
    ) -> "BIDSTable":
        """
        Filter the rows of the table.

        Args:
            key: Column to filter. Can be a metadata field, BIDS entity name, or any
                unprefixed column label in the `flat` table.
            value: Keep rows with this exact value.
            items: Keep rows whose value is in `items`.
            contains: Keep rows whose value contains `contains` (string only).
            regex: Keep rows whose value matches `regex` (string only).
            func: Apply an arbitrary function and keep values that evaluate to `True`.

        Returns:
            A filtered BIDS table.

        Example::
            filtered = (
                tab
                .filter("task", "rest")
                .filter("sub", items=["04", "06"])
                .filter("RepetitionTime", 2.0)
            )
        """
        # NOTE: Should be careful about reinventing a new style of query API. There are
        # some obvious things this can't do:
        #   - comparison operators <, >, <=, >=
        #   - negation
        #   - combining filters with 'or' instead of 'and'
        # At the bottom of this rabbit hole are more general query interfaces like those
        # already implemented in pandas, duckdb, polars. The goal should be not to
        # create a new one, but to make the 95% of use cases as easy as possible, and
        # empower users to interact with the underlying table using their more powerful
        # tool of choice if necessary.
        if sum(k is not None for k in [value, items, contains, regex, func]) != 1:
            raise ValueError(
                "Exactly one of value, items, contains, regex, or func must not be None"
            )

        try:
            # JSON metadata field
            # NOTE: Assuming all JSON metadata fields are uppercase.
            if key[:1].isupper():
                col = self.flat_meta[key]
            # Long name entity
            elif key in ENTITY_NAMES_TO_KEYS:
                col = self.ent[ENTITY_NAMES_TO_KEYS[key]]
            # Any other unprefixed column
            else:
                col = self.flat[key]
        except KeyError as exc:
            raise KeyError(
                f"Invalid key {key}; expected a valid BIDS entity or metadata field "
                "present in the dataset"
            ) from exc

        if value is not None:
            mask = col == value
        elif items is not None:
            mask = col.isin(items)
        elif contains is not None:
            mask = col.str.contains(contains)
        elif regex is not None:
            mask = col.str.match(regex)
        else:
            mask = col.apply(func)
        mask = mask.fillna(False).astype(bool)

        return self.loc[mask]

    def filter_multi(self, **filters) -> "BIDSTable":
        """
        Apply multiple filters to the table sequentially.

        Args:
            filters: A mapping of column labels to queries. Each query can either be
                a single value for an exact equality check or a `dict` for a more
                complex query, e.g. `{"items": [1, 2, 3]}`, that's passed through to
                `filter`.

        Returns:
            A filtered BIDS table.

        Example::
            filtered = tab.filter_multi(
                task="rest"
                sub={"items": ["04", "06"]},
                RepetitionTime=2.0,
            )
        """
        tab = self.copy(deep=False)

        for k, query in filters.items():
            if not isinstance(query, dict):
                query = {"value": query}
            tab = tab.filter(k, **query)
        return tab

    def sort_entities(
        self, by: Union[str, List[str]], inplace: bool = False
    ) -> "BIDSTable":
        """
        Sort the values of the table by entities.

        Args:
            by: label or list of labels. Can be `"dataset"` or a short or long entity
                name.
            inplace: sort the table in place

        Returns:
            A sorted BIDS table.
        """
        if isinstance(by, str):
            by = [by]

        # TODO: what about sorting by other columns, e.g. file_path?
        def add_prefix(k: str):
            if k == "dataset":
                k = f"ds__{k}"
            elif k in ENTITY_NAMES_TO_KEYS:
                k = f"ent__{ENTITY_NAMES_TO_KEYS[k]}"
            else:
                k = f"ent__{k}"
            return k

        by = [add_prefix(k) for k in by]
        out = self.sort_values(by, inplace=inplace)
        if inplace:
            return self
        return out

    def with_meta(self, inplace: bool = False) -> "BIDSTable":
        """
        Returns a new BIDS table complete with JSON sidecar metadata.
        """
        out = self if inplace else self.copy()
        file_paths = out.finfo["file_path"]
        meta_json = file_paths.apply(lambda path: extract_metadata(path)["json"])
        out.loc[:, "meta__json"] = meta_json
        return out

    @classmethod
    def from_df(cls, df: pd.DataFrame) -> "BIDSTable":
        """
        Create a BIDS table from a pandas `DataFrame` generated by `bids2table`.
        """
        return cls(df)

    @classmethod
    def from_parquet(cls, path: Path) -> "BIDSTable":
        """
        Read a BIDS table from a Parquet file or dataset directory generated by
        `bids2table`.
        """
        df = pd.read_parquet(path)
        return cls.from_df(df)

    @property
    def _constructor(self):
        # Makes sure that dataframe slices return a subclass instance
        # https://pandas.pydata.org/docs/development/extending.html#override-constructor-properties
        return BIDSTable


@dataclass
class BIDSFile:
    """
    A structured BIDS file.
    """

    dataset: str
    """Parent BIDS dataset."""
    root: Path
    """Path to parent dataset."""
    path: Path
    """File path."""
    entities: BIDSEntities
    """BIDS entities."""
    metadata: Dict[str, Any] = field(default_factory=dict)
    """BIDS JSON metadata."""

    @property
    def relative_path(self) -> Path:
        """
        The file path relative to the dataset root.
        """
        return self.path.relative_to(self.root)


def flat_to_multi_columns(df: pd.DataFrame, sep: str = "__") -> pd.DataFrame:
    """
    Convert a flat column index to a MultiIndex by splitting on `sep`.
    """
    # Do nothing if already a MultiIndex
    if isinstance(df.columns, pd.MultiIndex):
        return df

    # Do nothing for empty df
    # TODO: It would probably be better if the header was initialized even if there are
    # no records.
    if len(df.columns) == 0:
        return df

    split_columns = [col.split(sep) for col in df.columns]
    num_levels = max(map(len, split_columns))

    def _pad_col(col):
        return tuple((num_levels - len(col)) * [None] + col)

    df = df.copy(deep=False)
    df.columns = pd.MultiIndex.from_tuples(map(_pad_col, split_columns))
    return df


def multi_to_flat_columns(df: pd.DataFrame, sep: str = "__") -> pd.DataFrame:
    """
    Convert a column MultiIndex to a flat index by joining on `sep`.
    """
    # Do nothing if already flat
    if not isinstance(df.columns, pd.MultiIndex):
        return df

    columns = df.columns.to_flat_index()
    join_columns = [sep.join(col) for col in columns]

    df = df.copy(deep=False)
    df.columns = pd.Index(join_columns)
    return df


def join_bids_path(
    row: Union[pd.Series, Dict[str, Any]],
    prefix: Optional[Union[str, Path]] = None,
    valid_only: bool = True,
) -> Path:
    """
    Reconstruct a BIDS path from a table row or entities dict.

    Args:
        row: row from a `BIDSTable` or `BIDSTable.ent` subtable.
        prefix: output file prefix path.
        valid_only: only include valid BIDS entities.

    Example::

        tab = BIDSTable.from_parquet("dataset/index.b2t")
        paths = tab.apply(join_bids_path, axis=1)
    """
    # Filter in case input is a row from the raw dataframe and not the entities group.
    row = _filter_row(row, group="ent")
    entities = BIDSEntities.from_dict(row, valid_only=valid_only)
    path = entities.to_path(prefix=prefix, valid_only=valid_only)
    return path


def _filter_row(
    row: Union[pd.Series, Dict[str, Any]], group: str, sep: str = "__"
) -> Dict[str, Any]:
    """
    Filter a table row for fields from a particular group. Keeps all fields without a
    group prefix.
    """
    prefix = f"{group}{sep}"
    return {
        _removeprefix(k, prefix): v
        for k, v in row.items()
        if k.startswith(prefix) or sep not in k
    }


def _removeprefix(s: str, prefix: str) -> str:
    # same as str.removeprefix(), which was introduced in 3.9
    if s.startswith(prefix):
        s = s[len(prefix) :]
    return s



"""
A structured representation for BIDS entities.
"""

import re
import warnings
from dataclasses import asdict, dataclass, field, fields, make_dataclass
from functools import lru_cache
from pathlib import Path
from types import MappingProxyType
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import bidsschematools.schema
import pandas as pd
from elbow.typing import StrOrPath
from typing_extensions import Self, get_args, get_origin

_bids_schema_path: Optional[str] = None


def set_bids_schema_path(path: Optional[str]):
    """
    Set the path to the BIDS schema.
    """
    global _bids_schema_path
    _bids_schema_path = path


def _get_bids_schema() -> Any:
    """
    Get the BIDS schema.
    """
    global _bids_schema_path
    return bidsschematools.schema.load_schema(schema_path=_bids_schema_path)


BIDS_DATATYPES = tuple(o.value for o in _get_bids_schema().objects.datatypes.values())


def bids_field(
    name: str,
    display_name: str,
    required: bool = False,
    allowed_values: Optional[Iterable] = None,
    default: Optional[Any] = None,
    default_factory: Optional[Callable] = None,
):
    """
    BIDS entity dataclass field.
    """
    if allowed_values is not None:
        allowed_values = list(allowed_values)

    metadata = {
        "name": name,
        "display_name": display_name,
        "allowed_values": allowed_values,
    }
    if required:
        fld = field(metadata=metadata)
    elif default_factory is not None:
        fld = field(default_factory=default_factory, metadata=metadata)
    else:
        fld = field(default=default, metadata=metadata)
    return fld


@dataclass
class _BIDSEntitiesBase:
    """
    A dataclass representing known BIDS entities.

    Note:
        Only minimal validation of the entity values is performed. Use `bids_validator`
        for thorough validation of BIDS datasets.

    Reference:
        https://bids-specification.readthedocs.io/en/stable/appendices/entities.html
    """

    sub: str = bids_field(name="subject", display_name="Subject", required=True)
    ses: Optional[str] = bids_field(name="session", display_name="Session")

    datatype: Optional[str] = bids_field(
        name="datatype", display_name="Data type", allowed_values=BIDS_DATATYPES
    )
    suffix: Optional[str] = bids_field(name="suffix", display_name="Suffix")
    ext: Optional[str] = bids_field(name="extension", display_name="Extension")
    extra_entities: Optional[Dict[str, Union[str, int]]] = bids_field(
        name="extra_entities",
        display_name="Extra entities",
        default_factory=dict,
    )

    @staticmethod
    def special() -> List[str]:
        """
        Get list of field keys which are not standard entities.
        """
        return ["datatype", "suffix", "ext", "extra_entities"]

    @classmethod
    def from_dict(cls, entities: Dict[str, Any], valid_only: bool = False):
        """
        Initialize from a dict of entities.
        """
        filtered = {}
        extra_entities: Dict[str, Union[str, int]] = {}
        fields_map = {f.name: f for f in fields(cls) if f.name != "extra_entities"}

        def add_to_extra(k: Any, v: Any):
            if not isinstance(key, str):
                raise TypeError(
                    f"Extra entity {k} has type {type(k)}; only str supported"
                )
            if isinstance(v, (str, int)):
                extra_entities[k] = v
            else:
                warnings.warn(
                    f"Value {v} for extra entity {k} has type {type(v)}; "
                    f"only str, int supported"
                )

        for key, val in entities.items():
            if pd.isna(val):
                continue

            if key in fields_map:
                fld = fields_map[key]
                typ = _get_type(fld.type)
                try:
                    val = typ(val)
                except ValueError as exc:
                    raise ValueError(
                        f"Unable to coerce {repr(val)} to type {typ} for "
                        f"entity {fld.name}"
                    ) from exc

                allowed_values = fld.metadata.get("allowed_values")
                if allowed_values and val not in allowed_values:
                    raise ValueError(
                        f"Value {val} for entity {fld.name} isn't one of the "
                        f"allowed values {allowed_values}"
                    )

                filtered[key] = val

            # Special handling if the dict already contains 'extra_entities'. This makes
            # it easy to reconstruct entities from a df row.
            elif key == "extra_entities":
                assert isinstance(
                    val, dict
                ), "Value for 'extra_entities' key must be a dict"
                for k, v in val.items():
                    add_to_extra(k, v)

            elif not valid_only:
                add_to_extra(key, val)

        return cls(**filtered, extra_entities=extra_entities)

    @classmethod
    def from_path(cls, path: StrOrPath):
        """
        Initialize from a file path.
        """
        entities = parse_bids_entities(path)
        return cls.from_dict(entities)

    def to_dict(self, valid_only: bool = False) -> Dict[str, Any]:
        """
        Convert entities to a plain dict.
        """
        data = asdict(self)
        extra = data.pop("extra_entities")
        if not valid_only and extra:
            data.update(extra)
        return data

    def to_path(
        self,
        prefix: Optional[StrOrPath] = None,
        valid_only: bool = False,
        int_format: str = "%d",
    ) -> Path:
        """
        Generate a filepath based on the entitities.
        """
        special = {"datatype", "suffix", "ext"}
        data = self.to_dict(valid_only=valid_only)

        name = "_".join(
            _fmt_ent(k, v, int_format=int_format)
            for k, v in data.items()
            if k not in special and not pd.isna(v)
        )
        if self.suffix:
            name += f"_{self.suffix}"
        if self.ext:
            name += self.ext

        path = Path(name)
        if self.datatype:
            path = self.datatype / path
        if self.ses:
            path = f"ses-{self.ses}" / path
        path = f"sub-{self.sub}" / path
        if prefix:
            path = prefix / path
        return path

    def with_update(
        self, entitities: Optional[Dict[str, Any]] = None, **kwargs
    ) -> Self:
        """
        Create a new instance with updated entities.
        """
        data = self.to_dict(valid_only=False)
        if entitities:
            data.update(entitities)
        if kwargs:
            data.update(kwargs)
        return self.__class__.from_dict(data)


def make_bids_field(
    entity_schema: Dict[str, Any],
) -> Tuple[str, Any, Any]:
    """
    BIDS entity dataclass field.
    """
    metadata = {
        "name": entity_schema["name"],
        "display_name": entity_schema["display_name"],
        "allowed_values": entity_schema.get("enum"),
    }
    type_ = {"index": int}.get(entity_schema["format"], str)
    field_ = field(default=None, metadata=metadata)
    return entity_schema["name"], Optional[type_], field_


_sorted_entities = [
    _get_bids_schema().objects.entities[field_name]
    for field_name in _get_bids_schema().rules.entities
]
BIDSEntities: Any = make_dataclass(
    "BIDSEntities",
    [
        make_bids_field(dict(f))
        for f in _sorted_entities
        if f.name not in {f.name for f in fields(_BIDSEntitiesBase)}
    ],
    bases=(_BIDSEntitiesBase,),
)
del _sorted_entities


def _get_type(alias: Any) -> type:
    """
    Unbox type aliases of the form `Optional[str]`.
    """
    if _is_optional(alias):
        return get_args(alias)[0]
    return alias


def _is_optional(alias: Any) -> bool:
    """
    Check if alias is an optional of a primitive type like `Optional[str]`.
    """
    return (
        get_origin(alias) is Union
        and len(get_args(alias)) == 2
        and isinstance(None, get_args(alias)[1])
    )


def _fmt_ent(
    k: str,
    v: Union[str, int],
    *,
    int_format: str = "%d",
):
    if isinstance(v, int):
        v = int_format % v
    ent = f"{k}-{v}" if len(v) > 0 else k
    return ent


@lru_cache()
def parse_bids_entities(path: StrOrPath) -> Dict[str, str]:
    """
    Parse all BIDS filename ``"{key}-{value}"`` entities as well as special entities:

        - datatype
        - suffix
        - ext (extension)

    from the file path.

    .. note:: This function does not validate entities.
    """
    path = Path(path)
    entities = {}

    # datatype
    match = re.search(
        f"/({'|'.join(BIDS_DATATYPES)})/",
        path.as_posix(),
    )
    datatype = match.group(1) if match is not None else None

    filename = path.name
    parts = filename.split("_")

    # suffix and extension
    suffix_ext = parts.pop()
    idx = suffix_ext.find(".")
    if idx < 0:
        suffix, ext = suffix_ext, None
    else:
        suffix, ext = suffix_ext[:idx], suffix_ext[idx:]

    # suffix is actually an entity, put back in list
    if "-" in suffix:
        parts.append(suffix)
        suffix = None

    # parse entities
    for part in parts:
        if "-" in part:
            key, val = part.split("-", maxsplit=1)
        else:
            key, val = part, ""
        entities[key] = val

    for k, v in zip(["datatype", "suffix", "ext"], [datatype, suffix, ext]):
        if v is not None:
            entities[k] = v
    return entities


ENTITY_NAMES_TO_KEYS = MappingProxyType(
    {f.metadata["name"]: f.name for f in fields(BIDSEntities)}
)



import logging
from functools import partial
from pathlib import Path
from typing import List, Optional

from elbow.builders import build_parquet, build_table
from elbow.sources.filesystem import Crawler
from elbow.typing import StrOrPath

from CPACqc.bids2table.extractors.bids import extract_bids_subdir
from CPACqc.bids2table.table import BIDSTable

logger = logging.getLogger("bids2table")


def bids2table(
    root: StrOrPath,
    *,
    with_meta: bool = True,
    persistent: bool = False,
    index_path: Optional[StrOrPath] = None,
    exclude: Optional[List[str]] = None,
    incremental: bool = False,
    overwrite: bool = False,
    workers: Optional[int] = None,
    worker_id: Optional[int] = None,
    return_table: bool = True,
    subject: Optional[List[str]] = None,
) -> Optional[BIDSTable]:
    """
    Index a BIDS dataset directory and load as a pandas DataFrame.

    Args:
        root: path to BIDS dataset
        with_meta: extract JSON sidecar metadata. Excluding metadata can result in much
            faster indexing.
        persistent: whether to save index to disk as a Parquet dataset
        index_path: path to BIDS Parquet index to generate or load. Defaults to `root /
            "index.b2t"`. Index generation requires `persistent=True`.
        exclude: Optional list of directory names or glob patterns to exclude from indexing.
        incremental: update index incrementally with only new or changed files.
        overwrite: overwrite previous index.
        workers: number of parallel processes. If `None` or 1, run in the main
            process. Setting to <= 0 runs as many processes as there are cores
            available.
        worker_id: optional worker ID to use when scheduling parallel tasks externally.
            Specifying the number of workers is required in this case. Incompatible with
            overwrite.
        return_table: whether to return the BIDS table or just build the persistent
            index.
        subject: optional subject label to index only a specific subject directory

    Returns:
        A `BIDSTable` representing the indexed dataset(s), or `None` if `return_table`
        is `False`.
    """
    if not (return_table or persistent):
        raise ValueError("persistent and return_table should not both be False")

    root = Path(root).expanduser().resolve()
    if not root.is_dir():
        raise FileNotFoundError(f"root directory {root} does not exists")

    if exclude is None:
        exclude = []

    if subject is not None:
        subjects = [sub.lstrip("sub-") if sub.startswith("sub-") else sub for sub in subject]
        logger.info(f"Indexing only subjects: {', '.join([f'sub-{sub}' for sub in subjects])}")
        all_subjects = [d.name for d in root.iterdir() if d.is_dir() and d.name.startswith("sub-")]
        exclude += [f"{sub}" for sub in all_subjects if sub not in [f"sub-{s}" for s in subjects]]
        logger.info(f"Excluding subjects: {exclude}")
        include_patterns = [f"sub-{sub}" for sub in subjects]
        source = Crawler(
            root=root,
            include=include_patterns,
            exclude=exclude,
            dirs_only=True,
            follow_links=True,
        )
    else:
        logger.info("Indexing all subjects")
        source = Crawler(
            root=root,
            include=["sub-*"],  # find subject dirs
            skip=["sub-*"] + exclude,  # exclude specified patterns
            dirs_only=True,
            follow_links=True,
        )
    extract = partial(extract_bids_subdir, exclude=exclude, with_meta=with_meta)

    if index_path is None:
        index_path = root / "index.b2t"
    else:
        index_path = Path(index_path).resolve()

    stale = overwrite or incremental or worker_id is not None
    if index_path.exists() and not stale:
        if return_table:
            logger.info("Loading cached index %s", index_path)
            tab = BIDSTable.from_parquet(index_path)
        else:
            logger.info("Found cached index %s; nothing to do", index_path)
            tab = None
        return tab

    if not persistent:
        logger.info("Building index in memory")
        df = build_table(
            source=source,
            extract=extract,
            workers=workers,
            worker_id=worker_id,
        )
        tab = BIDSTable.from_df(df)
        return tab

    logger.info("Building persistent Parquet index")
    build_parquet(
        source=source,
        extract=extract,
        output=index_path,
        incremental=incremental,
        overwrite=overwrite,
        workers=workers,
        worker_id=worker_id,
        path_column="file__file_path",
        mtime_column="file__mod_time",
    )
    tab = BIDSTable.from_parquet(index_path) if return_table else None
    return tab


import logging
from typing import Any, Dict, Tuple

import nibabel as nib
import numpy as np
from elbow.record import Record
from elbow.typing import StrOrPath
from nibabel.filebasedimages import ImageFileError

from CPACqc.bids2table.entities import parse_bids_entities

try:
    import nifti

    has_nifti = True
except ModuleNotFoundError:
    has_nifti = False

logger = logging.getLogger(__name__)

# TODO: add more
IMAGE_EXTENSIONS = {".nii", ".nii.gz"}


def extract_image_meta(path: StrOrPath, *, backend: str = "nibabel") -> Record:
    entities = parse_bids_entities(path)
    ext = entities.get("ext")

    header = affine = None
    if ext in IMAGE_EXTENSIONS:
        try:
            header, affine = _read_image_meta(str(path), backend=backend)
        except (ImageFileError, SystemError) as exc:
            logger.warning("Failed to load image %s", path, exc_info=exc)

    rec = Record(
        {"image_header": header, "image_affine": affine},
        types={"image_header": "json", "image_affine": "ndarray<float64>"},
    )
    return rec


def _read_image_meta(
    path: str, backend: str = "nibabel"
) -> Tuple[Dict[str, Any], np.ndarray]:
    header: Dict[str, Any]
    affine: np.ndarray

    if backend == "nifti":
        if not has_nifti:
            raise ModuleNotFoundError("nifti image backend not installed")

        # TODO: the nifti C lib prints a lot of error output that I'd like to suppress
        header = nifti.read_header(path)
        # TODO: affine not currently implemented for nifti lib
        affine = None
    else:
        img = nib.load(path)
        if not isinstance(img, nib.Nifti1Image):
            raise TypeError(
                f"Foung image type {type(img).__name__}; only Nifti1Image supported"
            )

        header = {k: v for k, v in img.header.items()}
        affine = np.asarray(img.affine)

    header = {k: _cast_header_value(v) for k, v in header.items()}
    return header, affine


def _cast_header_value(value: np.ndarray) -> Any:
    """
    NiBabel header values appear to be numpy arrays whose data can be:

        - numeric vectors
        - numeric scalars
        - bytes scalars

    Convert these data to JSON-serializable python objects.

    TODO: Check these assumptions against more nibabel image types.
    """
    cast_value = value.tolist()
    if isinstance(cast_value, bytes):
        cast_value = cast_value.decode()
    return cast_value



from CPACqc.table.utils import *
from CPACqc.logging.log import logger

import os
import pandas as pd

def preprocess(df):
    for col in df.columns:
        if isinstance(df[col].iloc[0], dict):
            df[col] = df[col].apply(lambda x: str(x) if x else "")
            if df[col].nunique() == 1 and df[col].iloc[0] == "":
                df = df.drop(columns=[col])
    

    # Fill all columns with NaN with empty string
    df = df.fillna("")

    # Drop json column if it exists
    if "json" in df.columns:
        df = df.drop(columns=["json"])

    # Filter rows where file_path ends with .nii.gz
    nii_gz_files = df[df.file_path.str.endswith(".nii.gz")].copy()

    # Add a column that breaks the file_path to the last name of the file and drops extension
    nii_gz_files.loc[:, "file_name"] = nii_gz_files.file_path.apply(lambda x: os.path.basename(x).replace(".nii.gz", ""))

    nii_gz_files.loc[:, "resource_name"] = nii_gz_files.apply(gen_resource_name, axis=1)

    nii_gz_files = nii_gz_files[nii_gz_files.file_path.apply(lambda x: is_3d_or_4d(x))]

    # Check if the space column is empty and fill it accordingly
    nii_gz_files.loc[:, "space"] = nii_gz_files.apply(lambda x: fill_space(x), axis=1)

    # Combine sub and ses columns to create a new column called sub_ses
    nii_gz_files.loc[:, "sub_ses"] = nii_gz_files.apply(get_sub_ses, axis=1)

    # Create a new column called scan that combines task and run columns
    nii_gz_files.loc[:, "scan"] = nii_gz_files.apply(get_scan, axis=1)

    return nii_gz_files


import pandas as pd
import os
import argparse
from multiprocessing import Pool
from tqdm import tqdm
from functools import lru_cache
import tempfile

import nibabel as nib
from nibabel.orientations import io_orientation, ornt2axcodes
from colorama import Fore, Style, init

from CPACqc.plotting.plot import run
from CPACqc.bids2table._b2t import bids2table
from CPACqc.logging.log import logger

import json
import re

def get_file_info(file_path):
    img = nib.load(file_path)
    resolution = tuple(float(x) for x in img.header.get_zooms())
    dimension = tuple(int(x) for x in img.shape)
    
    affine = img.affine
    orientation =  "".join(ornt2axcodes(io_orientation(affine))) + " @nibabel"
    
    if len(dimension) == 4:
        # get TR info
        tr = float(img.header.get_zooms()[3])
        nos_tr = str(int(img.shape[-1]))
    else:
        tr = None
        nos_tr = None

    return json.dumps({
        "resolution": resolution,
        "dimension": dimension,
        "tr": tr,
        "nos_tr": nos_tr,
        "orientation": orientation
    })

def gen_resource_name(row):
    sub = row["sub"]
    ses = row["ses"] if row["ses"] != "" else False

    sub_ses = f"sub-{sub}_ses-{ses}" if ses else f"sub-{sub}"

    task = row["task"] if row["task"] != "" else False
    run = row["run"] if row["run"] != "" else False
    
    # Create a flexible pattern for the scan part
    scan = f"task-{task}_run-\\d*_" if task and run else ""
    
    # Use regular expression to replace the pattern
    pattern = re.escape(f"{sub_ses}_") + scan
    resource_name = re.sub(pattern, "", row["file_name"])
    
    return resource_name

# add a utility function to return rows provided a resource_name
def get_rows_by_resource_name(resource_name, nii_gz_files):
    # Ensure nii_gz_files is a DataFrame and access the correct column
    if isinstance(nii_gz_files, pd.DataFrame):
        rows = nii_gz_files[nii_gz_files['resource_name'].str.endswith(resource_name)]
        if len(rows) == 0:
            logger.error(f"NOT FOUND: {resource_name}")
            return None
        return rows
    else:
        logger.error("nii_gz_files is not a DataFrame")
        return None

# check file_path and drop the ones that are higher dimensions for now
def is_3d_or_4d(file_path):
    dim = len(nib.load(file_path).shape)
    if dim > 4:
        file_name = os.path.basename(file_path).split(".")[0]
        logger.error(f"NOT 3D: {file_name} \n its {dim}D")
        logger.error(f"Skipping for now ....")
        return False
    return True

def get_scan(row):
    task = row["task"] if row["task"]  else False
    run = int(row["run"]) if row["run"]  else False
    return f"task-{task}_run-{run}" if task and run else ""

def get_sub_ses(row):
    sub = row["sub"]
    ses = row["ses"] if row["ses"] != "" else False
    return f"sub-{sub}_ses-{ses}" if ses else f"sub-{sub}"

def gen_filename(res1_row, res2_row=None):
    scan = f"task-{res1_row['task']}_run-{int(res1_row['run'])}_" if res1_row['task'] and res1_row['run'] else ""
    if res2_row is not None:
        return f"sub-{res1_row['sub']}_ses-{res1_row['ses']}_{scan + res1_row['resource_name']} overlaid on {res2_row['resource_name']}"
    else:
        return f"sub-{res1_row['sub']}_ses-{res1_row['ses']}_{scan + res1_row['resource_name']}"

def create_directory(sub, ses, base_dir):
    sub_dir = os.path.join(base_dir, sub, ses)
    os.makedirs(sub_dir, exist_ok=True)
    return sub_dir

def generate_plot_path(sub_dir, file_name):
    return os.path.join(sub_dir, f"{file_name}.png")

def process_row(row, nii_gz_files, overlay_dir, plots_dir):
    image_1 = row.get("image_1", False)
    image_2 = row.get("image_2", False)

    resource_name_1 = get_rows_by_resource_name(image_1, nii_gz_files) if image_1 else None
    resource_name_2 = get_rows_by_resource_name(image_2, nii_gz_files) if image_2 else None

    if resource_name_1 is None:
        logger.error(f"NOT FOUND: {image_1}")
        return []

    result_rows = []
    seen = set()  # To track duplicates

    for _, res1_row in resource_name_1.iterrows():
        if resource_name_2 is not None:
            result_rows.extend(process_res2_rows(res1_row, resource_name_2, seen, overlay_dir, plots_dir))
        else:
            result_rows.extend(process_single_row(res1_row, seen, plots_dir))

    return result_rows

def process_res2_rows(res1_row, resource_name_2, seen, overlay_dir, plots_dir):
    result_rows = []
    for _, res2_row in resource_name_2.iterrows():
        file_name = gen_filename(res1_row, res2_row)
        if res1_row['space'] == res2_row['space']:
            if file_name not in seen:
                seen.add(file_name)
                sub_dir = create_directory(res1_row['sub'], res1_row['ses'], overlay_dir)
                plot_path = generate_plot_path(sub_dir, file_name)
                result_rows.append(create_result_row(res1_row, res2_row, file_name, overlay_dir, plot_path))
        else:
            logger.info(f"SPACE MISMATCH while trying to plot .. \n{file_name.split('overlaid on')[0]}\nover\n{file_name.split(' overlaid on ')[1]}\n{res1_row['space']} != {res2_row['space']}")
    if len(result_rows) == 0:
        logger.error(f"MATCHING SPACE NOT-FOUND to plot {res1_row['resource_name']} over {resource_name_2['resource_name'].tolist()}")
        file_name = gen_filename(res1_row) + " *SPACE MISMATCH TO OVERLAY!*"
        if file_name not in seen:
            seen.add(file_name)
            sub_dir = create_directory(res1_row['sub'], res1_row['ses'], plots_dir)
            plot_path = generate_plot_path(sub_dir, file_name)
            result_rows.append(create_result_row(res1_row, None, file_name, plots_dir, plot_path))
    return result_rows

def process_single_row(res1_row, seen, plots_dir):
    result_rows = []
    file_name = gen_filename(res1_row)
    if file_name not in seen:
        seen.add(file_name)
        sub_dir = create_directory(res1_row['sub'], res1_row['ses'], plots_dir)
        plot_path = generate_plot_path(sub_dir, file_name)
        result_rows.append(create_result_row(res1_row, None, file_name, plots_dir, plot_path))
    return result_rows

def create_result_row(res1_row, res2_row, file_name, plots_dir, plot_path):
    return {
        "sub": res1_row["sub"],
        "ses": res1_row["ses"],
        "file_path_1": res1_row["file_path"],
        "file_path_2": res2_row["file_path"] if res2_row is not None else None,
        "file_name": file_name,
        "plots_dir": plots_dir,
        "plot_path": plot_path,
        "datatype": res1_row["datatype"],
        "resource_name": res1_row["resource_name"],
        "space": res1_row["space"],
        "scan": res1_row["scan"]
    }

def parse_bids(base_dir, sub=None, workers=8):
    print(Fore.YELLOW + "Parsing BIDS directory..." + Style.RESET_ALL)
    if logger: 
        logger.info("Parsing BIDS directory...")

    df = bids2table(base_dir, subject=sub, workers=workers).flat
    return df

def run_wrapper(args):
    return run(*args)

def fill_space(row):
    if row["space"] == "":
        if row["datatype"] == "anat":
            return "T1w"
        elif row["datatype"] == "func":
            return "bold"
    return row["space"]



import logging
import sys
from pathlib import Path
from typing import Dict, Optional, Tuple, Union


class RepetitiveFilter(logging.Filter):
    """
    Suppress similar log messages after a number of repeats.
    """

    def __init__(self, max_repeats: int = 5):
        self.max_repeats = max_repeats

        self._counts: Dict[Tuple[str, int], int] = {}

    def filter(self, record: logging.LogRecord):
        key = record.pathname, record.lineno
        count = self._counts.get(key, 0)
        if count == self.max_repeats:
            record.msg += " [future messages suppressed]"
        self._counts[key] = count + 1
        return count <= self.max_repeats


def setup_logging(
    name: Optional[str] = None,
    level: Optional[Union[int, str]] = None,
    log_path: Optional[Union[str, Path]] = None,
    max_repeats: Optional[int] = 5,
    overwrite: bool = False,
    propagate: bool = False,
) -> logging.Logger:
    """
    Setup a logger in a particular way.
    """
    logger = logging.getLogger(name=name)
    if len(logger.handlers) > 0 and not overwrite:
        return logger

    if level is not None:
        logger.setLevel(level)

    # clean up any pre-existing filters and handlers
    for f in logger.filters:
        logger.removeFilter(f)
    for h in logger.handlers:
        logger.removeHandler(h)
        h.close()

    if max_repeats:
        logger.addFilter(RepetitiveFilter(max_repeats=max_repeats))

    fmt = "[%(levelname)s %(name)s %(asctime)s]: %(message)s"
    formatter = logging.Formatter(fmt, datefmt="%y-%m-%d %H:%M:%S")

    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logger.level)
    logger.addHandler(stream_handler)

    if log_path is not None:
        file_handler = logging.FileHandler(log_path, mode="a")
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logger.level)
        logger.addHandler(file_handler)

    if not propagate:
        logger.propagate = False
    return logger



import argparse
import sys
from pathlib import Path

from CPACqc.bids2table import bids2table
from CPACqc.bids2table.logging import setup_logging


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("root", metavar="ROOT", type=Path, help="Path to BIDS dataset")
    parser.add_argument(
        "--output",
        "-o",
        type=Path,
        help="Path to output parquet dataset directory (default: {ROOT}/index.b2t)",
        default=None,
    )
    parser.add_argument(
        "--incremental",
        "--inc",
        help="Update index incrementally with only new or changed files.",
        action="store_true",
    )
    parser.add_argument(
        "--overwrite",
        "-x",
        help="Overwrite previous index.",
        action="store_true",
    )
    parser.add_argument(
        "--workers",
        "-w",
        metavar="COUNT",
        type=int,
        help="Number of worker processes. Setting to -1 runs as many processes as "
        "there are cores available. (default: 1)",
        default=1,
    )
    parser.add_argument(
        "--worker_id",
        "--id",
        metavar="RANK",
        type=int,
        help="Optional worker ID to use when scheduling parallel tasks externally. "
        "Incompatible with --overwrite. (default: None)",
        default=None,
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="Increase verbosity level.",
    )
    parser.add_argument(
        "--exclude",
        nargs="+",
        default=None,
        help="List of directory names or glob patterns to exclude from indexing.",
    )
    parser.add_argument(
        "--subject",
        "-sub",
        nargs="+",
        type=str,
        help="List of subject labels to index (default: None)",
        default=None,
    )
    args = parser.parse_args()

    log_level = ["ERROR", "WARNING", "INFO"][min(args.verbose, 2)]
    setup_logging(level=log_level)

    bids2table(
        root=args.root,
        persistent=True,
        index_path=args.output,
        incremental=args.incremental,
        overwrite=args.overwrite,
        workers=args.workers,
        worker_id=args.worker_id,
        exclude=args.exclude,
        subject=args.subject,
        return_table=False,
    )


if __name__ == "__main__":
    sys.exit(main())



import logging
import os

logger = None

def setup_logger():
    # setup logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    # Create a file handler
    log_file = os.path.join(os.getcwd(), 'CPACqc.log')
    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.INFO)

    # Create a logging format
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)

    # Add the handlers to the logger
    logger.addHandler(file_handler)
    
    return logger

logger = setup_logger()


from functools import lru_cache
from pathlib import Path
from typing import Dict, Generator, List, Optional

from elbow.typing import StrOrPath

from CPACqc.bids2table.entities import parse_bids_entities

from .dataset import is_dataset_root


def find_bids_parents(
    query: Dict[str, str],
    start: StrOrPath,
    depth: Optional[int] = None,
) -> Generator[str, None, None]:
    """
    Find all BIDS files satisfying the inheritance principle requirements for the given
    `query` entities dict. The `query` must contain at least one of `'suffix'` or
    `'ext'`. Search up the directory hierarchy at most `depth` levels, starting from and
    including `start`, or until a `dataset_description.json` file is found, indicating
    the BIDS dataset root directory. If `depth` is `None`, the search may continue to
    the filesystem root.

    Yields matching `path`s in decreasing topological order.
    """
    suffix = query.get("suffix")
    ext = query.get("ext")
    if not (suffix or ext):
        raise ValueError("At least one of 'suffix' or 'ext' are required in `query`.")
    pattern = f"*{suffix}{ext}" if suffix else f"*{ext}"

    start = Path(start).absolute()
    if not start.is_dir():
        start = start.parent

    if depth is None:
        depth = len(start.parts)

    for _ in range(depth):
        for path in _glob(start, pattern):
            entities = parse_bids_entities(path)
            if _test_bids_match(query, entities):
                yield str(path)

        # Stop climbing the directory if we find the description json, which should
        # always be at the top-level dataset directory.
        # TODO: for nested datasets, can you inherit beyond the first root? I hope not..
        if is_dataset_root(start):
            break

        start = start.parent


def find_first_bids_parent(
    query: Dict[str, str], start: StrOrPath, depth: Optional[int] = None
) -> Optional[str]:
    """
    Find the first BIDS parent file matching the ``query`` entities dict. Returns
    ``None`` if no parents found. See :func:`find_bids_parents` for more details.
    """
    return next(find_bids_parents(query, start, depth), None)


@lru_cache()
def _glob(path: Path, pattern: str) -> List[Path]:
    return list(path.glob(pattern))


def _test_bids_match(query: Dict[str, str], entities: Dict[str, str]) -> bool:
    """
    Test if entities satisfies the inheritance principle for query.
    """
    entities = entities.copy()
    entities.pop("datatype", None)
    return set(entities).issubset(query) and all(
        query[k] == v for k, v in entities.items()
    )



from CPACqc.report.utils import *

from reportlab.lib.pagesizes import letter
from reportlab.pdfgen import canvas
from reportlab.lib.utils import ImageReader
from reportlab.platypus import Paragraph, Table, TableStyle
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib import colors

from colorama import Fore, Style
import pandas as pd
import os
import json


def make_pdf(df, qc_dir, sub_ses):
    print(Fore.YELLOW + "Generating PDF report..." + Style.RESET_ALL)

    # Handle .pdf in pdf_name
    pdf = f"{sub_ses}_qc_report.pdf"

    # Determine if pdf is a full path or just a file name
    if os.path.isabs(pdf):
        pdf_path = pdf
    else:
        pdf_path = os.path.join(os.getcwd(), pdf)

    c = canvas.Canvas(pdf_path, pagesize=letter)
    width, height = letter

    # Add CPAC logo and description to the front page
    logo_path = 'https://avatars.githubusercontent.com/u/2230402?s=200&v=4'  # Adjust the path as needed
    logo_img = ImageReader(logo_path)
    logo_width = 150  # Adjust the logo width
    logo_height = 150  # Adjust the logo height

    # Title at the top
    c.setFont("Helvetica", 30)
    c.drawCentredString(width / 2, height - 100, f"{sub_ses}")

    # Logo in the middle
    c.drawImage(logo_img, (width - logo_width) / 2, (height - logo_height) / 2, width=logo_width, height=logo_height)
    c.setFont("Helvetica", 15)
    c.drawCentredString(width / 2, height - 120, "Quality Control Report")

    # Footer information
    c.setFont("Helvetica", 12)
    c.drawCentredString(width / 2, 100, f"Created on: {pd.Timestamp.now().strftime('%Y-%m-%d %H:%M:%S')}")
    c.drawCentredString(width / 2, 80, "CPAC developers")

    # Add an initial page to skip the first page
    c.showPage()

    # Contents page
    c.setFont("Helvetica", 20)
    c.drawCentredString(width / 2, height - 100, "Contents")
    c.setFont("Helvetica", 12)
    y_position = height - 150
    contents = []

    # Get unique chapters and sort them in ascending order
    chapters = sorted(set(df['datatype'].dropna()))

    for chapter in chapters:
        contents.append(f"{chapter}")
        y_position -= 20
        c.setFillColor(colors.blue)
        c.drawString(100, y_position, f"{chapter}")
        c.linkRect("", f"chapter_{chapter}", (100, y_position - 2, 200, y_position + 10), color=colors.blue)
        y_position -= 10

        # List subsections under each chapter if they exist
        chapter_data = df[df['datatype'] == chapter]
        scans = sorted(set(chapter_data['scan'].dropna()))
        for scan in scans:
            c.setFillColor(colors.green)
            c.drawString(120, y_position, f"{scan}")
            c.linkRect("", f"subsection_{chapter}_{scan}", (120, y_position - 2, 220, y_position + 10), color=colors.green)
            y_position -= 20

    y_position = height - 30  # Start at the top of the new page
    page_number = 1  # Initialize page number

    # Add all images to the PDF, grouped by chapter and scan
    for chapter in chapters:
        # Add a new page for each chapter
        c.showPage()
        page_number += 1  # Increment page number
        y_position = height - 30  # Reset y_position for the new page

        # Add chapter name to the blank page
        c.setFont("Helvetica", 30)
        c.drawCentredString(width / 2, height / 2, chapter)
        c.bookmarkPage(f"chapter_{chapter}")

        chapter_data = df[df['datatype'] == chapter]
        scans = sorted(set(chapter_data['scan'].dropna()))

        for scan in scans:
            # Add a new page for each scan
            c.showPage()
            page_number += 1  # Increment page number
            y_position = height - 30  # Reset y_position for the new page

            # Add chapter name and subsection string at the top of each page
            c.setFont("Helvetica", 12)
            c.drawString(10, height - 30, f"Chapter: {chapter}/{scan}")
            c.bookmarkPage(f"subsection_{chapter}_{scan}")

            scan_data = chapter_data[chapter_data['scan'] == scan]

            if not scan_data.empty:
                for _, image_data in scan_data.iterrows():
                    image_path = os.path.join(qc_dir, image_data['relative_path'])
                    if os.path.exists(image_path):
                        img = ImageReader(image_path)
                        max_img_width = width - 20  # Adjust the max width to fit the page
                        max_img_height = height - 100  # Adjust the max height to fit the page

                        # Preserve aspect ratio
                        img_width, img_height = img.getSize()
                        aspect_ratio = img_width / img_height
                        if aspect_ratio > 1:
                            img_width = max_img_width
                            img_height = img_width / aspect_ratio
                        else:
                            img_height = max_img_height
                            img_width = img_height * aspect_ratio

                        # Check if the image fits on the current page, otherwise add a new page
                        if y_position - img_height - 140 < 0:  # Adjusted to account for additional text and white space
                            c.drawRightString(width - 30, 20, str(page_number))  # Add page number
                            c.showPage()
                            page_number += 1  # Increment page number
                            y_position = height - 30  # Reset y_position for the new page

                            # Add chapter name and subsection string at the top of each page
                            c.setFont("Helvetica", 12)
                            c.drawString(10, height - 30, f"Chapter : {chapter}/{scan}")
                            c.bookmarkPage(f"subsection_{chapter}_{scan}")

                        # Print the image name above each image
                        c.setFont("Helvetica", 15)
                        c.drawString(10, y_position - 20, f"{image_data['resource_name']}")

                        # Add the image to the PDF
                        c.drawImage(img, (width - img_width) / 2, y_position - img_height - 40, width=img_width, height=img_height)
                        
                        # Use Paragraph to wrap the label text
                        label = f"{image_data['file_name']}"
                        styles = getSampleStyleSheet()
                        styles['Normal'].textColor = colors.whitesmoke
                        wrapped_label = Paragraph(label, styles['Normal'])
                        wrapped_label.wrapOn(c, width - 20, height)

                        # Add file information under the image label
                        file_info = json.loads(image_data['file_info'])
                        file_info_text = [
                            ["Image:", wrapped_label],
                            ["Orientation:", file_info['orientation']],
                            ["Dimensions:", " x ".join(map(str, file_info['dimension']))],
                            ["Resolution:", " x ".join(map(lambda x: str(round(x, 2)), file_info['resolution']))]
                        ]

                        if file_info['tr'] is not None:
                            file_info_text.append(["TR:", str(round(file_info['tr'], 2))])

                        if file_info['nos_tr'] is not None:
                            file_info_text.append(["No of TRs:", str(file_info['nos_tr'])])

                        table = Table(file_info_text, colWidths=[80, width - 100])
                        table.setStyle(TableStyle([
                            ('BACKGROUND', (0, 0), (-1, 0), colors.darkgreen),
                            ('TEXTCOLOR', (0, 0), (-1, 0), colors.whitesmoke),
                            ('ALIGN', (0, 0), (-1, -1), 'LEFT'),
                            ('FONTNAME', (0, 0), (-1, 0), 'Helvetica-Bold'),
                            ('FONTSIZE', (0, 0), (-1, -1), 10),
                            ('BOTTOMPADDING', (0, 0), (-1, 0), 12),
                            ('BACKGROUND', (0, 1), (-1, -1), colors.lightgrey),
                            ('TEXTCOLOR', (0, 1), (-1, -1), colors.black),
                            ('GRID', (0, 0), (-1, -1), 1, colors.black),
                        ]))
                        table.wrapOn(c, width - 20, height)
                        table_height = table.wrap(width - 20, height)[1]
                        table.drawOn(c, 10, y_position - img_height - table_height - 60)

                        # Move to the next row after each image
                        y_position -= img_height  + table_height + 70  # Adjusted to account for additional text and white space

    # Add the final page number
    c.drawRightString(width - 30, 20, str(page_number))

    # Save the PDF
    c.save()
    print(Fore.GREEN + "PDF report generated successfully." + Style.RESET_ALL)


from concurrent.futures import ProcessPoolExecutor, as_completed
from tqdm import tqdm
from colorama import Fore, Style

from CPACqc.logging.log import logger

def run_multiprocessing(func, args, n_procs):
    not_plotted = []
    with ProcessPoolExecutor(max_workers=n_procs) as executor:
        futures = {executor.submit(func, arg): arg for arg in args}
        for future in tqdm(as_completed(futures), total=len(futures), desc="Processing ..."):
            try:
                result = future.result()
                logger.info(f"Successfully processed {futures[future]}: {result}")
            except Exception as e:
                if "terminated abruptly" in str(e):
                    print(Fore.RED + f"Error processing {futures[future]}: {e}\n Try with lower number of processes" + Style.RESET_ALL)
                logger.error(f"Error processing {futures[future]}: {e}, Try with a lower number of processes")
                not_plotted.append(futures[future])
    return not_plotted


r"""
[bids2table](https://github.com/childmindresearch/bids2table) is a library for efficiently
indexing and querying large-scale BIDS neuroimaging datasets and derivatives. It aims to
improve upon the efficiency of [PyBIDS](https://github.com/bids-standard/pybids) by
leveraging modern data science tools.

bids2table represents a BIDS dataset index as a single table with columns for BIDS
entities and file metadata. The index is constructed using
[Arrow](https://arrow.apache.org/) and stored in [Parquet](https://parquet.apache.org/)
format, a binary tabular file format optimized for efficient storage and retrieval.

## Installation

A pre-release version of bids2table can be installed with

```sh
pip install bids2table
```

The latest development version can be installed with

```sh
pip install git+https://github.com/childmindresearch/bids2table.git
```

## Quickstart

The main entrypoint to the library is the `bids2table.bids2table` function, which builds
the index.

```python
tab = bids2table("path/to/dataset")
```

You can also build the index in parallel

```python
tab = bids2table("path/to/dataset", workers=8)
```

To save the index to disk as a [Parquet](https://parquet.apache.org/) dataset for later
reuse, run

```python
tab = bids2table("path/to/dataset", persistent=True)
```

By default this saves the index to an `index.b2t` directory under the dataset root
directory. To change the output destination, use the `index_path` argument.

To generate and save an index from the command line, you can use the `bids2table` CLI.

```sh
usage: bids2table [-h] [--output OUTPUT] [--incremental] [--overwrite] [--workers COUNT]
                  [--worker_id RANK] [--verbose]
                  ROOT
```

See `bids2table --help` for more information.

## Table representation

The generated index is represented as a `bids2table.BIDSTable`, which is just a subclass
of a `pandas.DataFrame`. Each row in the table corresponds to a BIDS data file, and the
columns are organized into groups:

- dataset (`BIDSTable.ds`): dataset name, relative dataset path, and the JSON dataset
description
- entities (`BIDSTable.ent`): All [valid BIDS
entities](https://bids-specification.readthedocs.io/en/stable/appendices/entities.html)
plus an `extra_entities` dict containing any extra entities
- metadata (`BIDSTable.meta`): BIDS JSON metadata
- file info (`BIDSTable.finfo`): General file info including the full file path and last
modified time

The `BIDSTable` also makes it easy to access some of the key characteristics of your
dataset. The BIDS datatypes, modalities, subjects, and entities present in the dataset
are each accessible as properties of the table.

In addition, the associated JSON metadata for each file can be conveniently accessed via
the `BIDSTable.flat_meta` property.

### Filtering

To filter the table for files matching certain criteria, you can use the
`BIDSTable.filter` method which selects for rows based on whether a specified column
meets a condition.

```python
filtered = (
    tab
    .filter("task", "rest")
    .filter("sub", items=["04", "06"])
    .filter("RepetitionTime", 2.5)
)
```

To apply multiple filters at once, you can also use `BIDSTable.filter_multi`

```python

filtered = tab.filter_multi(
    task="rest"
    sub={"items": ["04", "06"]},
    RepetitionTime=2.5,
)
```

This is similar to the
[`BIDSLayout.get()`](https://bids-standard.github.io/pybids/generated/bids.layout.BIDSLayout.html#bids.layout.BIDSLayout)
method in [`PyBIDS`](https://bids-standard.github.io/pybids/), where each `key=value`
pair specifies the column to filter on and the condition to apply.

### Advanced usage

For more advanced usage that goes beyond what's supported in this higher-level
interface, you can also interact directly with the underlying
[`DataFrame`](https://pandas.pydata.org/docs/user_guide/index.html).

The column labels of the raw table indicate the group as a prefix, e.g. `ent__*` for
BIDS entities. However, you may find one of the alternative views of the table more
useful:

- `BIDSTable.nested`: Columns organized as a nested pandas
[`MultiIndex`](https://pandas.pydata.org/docs/user_guide/advanced.html#hierarchical-indexing-multiindex).
- `BIDSTable.flat`: Flattened columns without any nesting or group prefix.

.. warning::
    You should avoid manipulating the table in place if possible, as this may interfere
    with the higher-level accessors. If you must manipulate in place, consider
    converting the `BIDSTable` to a plain `DataFrame` first.

    ```python
    df = pd.DataFrame(tab)
    ```
"""

# # Register elbow extension types
# import elbow.dtypes  # noqa

# from ._b2t import bids2table
# #from ._version import __version__, __version_tuple__  # noqa
# from .entities import BIDSEntities, parse_bids_entities
# from .table import BIDSFile, BIDSTable, join_bids_path

# __all__ = [
#     "bids2table",
#     "BIDSTable",
#     "BIDSFile",
#     "BIDSEntities",
#     "parse_bids_entities",
#     "join_bids_path",
# ]



import pandas as pd
import os
import pandas as pd
import os
import argparse

from CPACqc.table.table import preprocess
from CPACqc.table.utils import *
from CPACqc.multiprocessing.multiprocessing_utils import run_multiprocessing
from CPACqc.report.pdf import make_pdf
from CPACqc.logging.log import logger

def main(bids_dir, qc_dir, config=False, sub=None, n_procs=8, pdf=False):
    os.makedirs(qc_dir, exist_ok=True)
    
    logger.info(f"Running QC with nprocs {n_procs}...")
    
    # Create necessary directories
    for directory in ["plots", "overlays", "csv"]:
        os.makedirs(os.path.join(qc_dir, directory), exist_ok=True)
    
    plots_dir = os.path.join(qc_dir, "plots")
    overlay_dir = os.path.join(qc_dir, "overlays")
    csv_dir = os.path.join(qc_dir, "csv")

    if sub and isinstance(sub, str):
        sub = [sub]

    df = parse_bids(bids_dir, sub=sub, workers=n_procs)
    
    nii_gz_files = preprocess(df)

    # split the df into different df based on unique sub_ses
    sub_ses = nii_gz_files["sub_ses"].unique()
    no_sub_ses = len(sub_ses)
    if no_sub_ses == 0:
        logger.error("No subjects found.")
        print(Fore.RED + "No subjects found." + Style.RESET_ALL)
        return

    not_plotted = []
    # different df for each sub_ses
    for index, sub_ses in enumerate(sub_ses):
        index = index + 1
        sub_df = nii_gz_files[nii_gz_files["sub_ses"] == sub_ses]

        print(Fore.YELLOW + f"Processing {sub_ses} ({index}/{no_sub_ses})..." + Style.RESET_ALL)
        logger.info(f"Processing {sub_ses} ({index}/{no_sub_ses})...")

        if config:
            overlay_df = pd.read_csv(config).fillna(False)
            results = overlay_df.apply(lambda row: process_row(row, sub_df, overlay_dir, plots_dir), axis=1).tolist()
            results = [item for sublist in results for item in sublist]  # Flatten the list of lists
            result_df = pd.DataFrame(results)
        else:
            result_df = sub_df.copy()
            result_df['file_path_1'] = sub_df['file_path']
            result_df['file_path_2'] = None
            result_df['file_name'] = result_df.apply(lambda row: gen_filename(res1_row=row), axis=1)
            result_df['plots_dir'] = plots_dir
            result_df['plot_path'] = result_df.apply(lambda row: generate_plot_path(create_directory(row['sub'], row['ses'], row['plots_dir']), row['file_name']), axis=1)
            result_df = result_df[['sub', 'ses', 'file_path_1', 'file_path_2', 'file_name', 'plots_dir', 'plot_path', 'datatype', 'resource_name', 'space', 'scan']].copy()

        result_df['relative_path'] = result_df.apply(lambda row: os.path.relpath(row['plot_path'], qc_dir), axis=1)
        result_df['file_info'] = result_df.apply(lambda row: get_file_info(row['file_path_1']), axis=1)
        
        result_df_csv_path = os.path.join(csv_dir, f"{sub_ses}_results.csv")
        result_df.to_csv(result_df_csv_path, mode='a' if os.path.exists(result_df_csv_path) else 'w', header=not os.path.exists(result_df_csv_path), index=False)
    
        args = [
            (
                row['sub'], 
                row['ses'],  
                row['file_path_1'],
                row['file_path_2'], 
                row['file_name'],
                row['plots_dir'],
                row['plot_path'],
            ) 
            for _, row in result_df.iterrows()
        ]

        not_plotted += run_multiprocessing(run_wrapper, args, n_procs)

        if pdf:
            try:
                make_pdf(result_df, qc_dir, sub_ses)
            except Exception as e:
                logger.error(f"Error generating PDF: {e}")
                print(Fore.RED + f"Error generating PDF: {e}" + Style.RESET_ALL)

    return not_plotted
    
if __name__ == "__main__":
    import argparse

    # Parse command-line arguments
    parser = argparse.ArgumentParser(description="Process BIDS directory and generate QC plots.")
    parser.add_argument("-d", "--bids_dir", required=True, help="Path to the BIDS directory")
    parser.add_argument("-o", "--qc_dir", required=True, help="Path to the QC output directory")
    parser.add_argument("-c", "--config", required=False, help="Config file")
    parser.add_argument("-s", "--sub", nargs='+', required=False, help="Specify subject/participant label(s) to process")
    parser.add_argument("-n", "--n_procs", type=int, default=8, help="Number of processes to use for multiprocessing")
    parser.add_argument("-v", "--version", action='version', version=f'%(prog)s {__version__}', help="Show the version number and exit")
    parser.add_argument("-pdf", "--pdf", required=False, help="Generate PDF report")

    args = parser.parse_args()
    main(args.bids_dir, args.qc_dir, args.config, args.sub, args.n_procs, args.pdf)


import json
import logging
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from elbow.record import Record
from elbow.typing import StrOrPath

logger = logging.getLogger(__name__)


def extract_dataset(path: StrOrPath) -> Record:
    """
    Get info about the BIDS dataset that ``path`` belongs to.
    """
    name, root = identify_bids_dataset(path)
    desc = get_dataset_description(root) if root is not None else None
    rec = Record(
        {
            "dataset": name,
            "dataset_type": desc.get("DatasetType") if desc else None,
            "dataset_path": str(root.absolute()) if root else None,
            "dataset_description": desc,
        },
        types={"dataset_type": str, "dataset_description": "json"},
    )
    return rec


def identify_bids_dataset(path: StrOrPath) -> Tuple[Optional[str], Optional[Path]]:
    """
    Identify the BIDS dataset that ``path`` belongs to. Return the dataset directory
    name and the full dataset path. For nested derivatives datasets, a composite name of
    the form ``"ds000001/derivatives/fmriprep"`` is returned.

    Note that the name is extracted from the path, not the dataset description JSON.
    """
    path = Path(path)
    parent = path if path.is_dir() else path.parent

    parts: List[str] = []
    scanning = False
    top_idx = None
    root = None

    while parent.name:
        if is_dataset_root(parent):
            scanning = True
            top_idx = len(parts)
            if root is None:
                root = parent

        if scanning:
            parts.append(parent.name)

        parent = parent.parent

    if len(parts) == 0:
        logger.warning("File %s is not part of any valid BIDS dataset.", path)
        return None, None

    parts = parts[: top_idx + 1]
    dataset = "/".join(reversed(parts))
    return dataset, root


@lru_cache(maxsize=512)
def is_dataset_root(path: Path) -> bool:
    """
    Test if ``path`` is a BIDS dataset root directory.
    """
    return path.is_dir() and (path / "dataset_description.json").exists()


@lru_cache(maxsize=64)
def get_dataset_description(root: Path) -> Optional[Dict[str, Any]]:
    """
    Load the JSON description for the BIDS dataset root directory ``root``.
    """
    if not is_dataset_root(root):
        raise ValueError(f"{root} is not a BIDS dataset root")

    desc_path = root / "dataset_description.json"
    with desc_path.open() as f:
        try:
            description = json.load(f)
        except json.JSONDecodeError:
            description = None
    return description
